/***********************************************************************************

	Projekt: 	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei: 		timer.h
	Autor: 		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Funktions-/Variablendeklarationen und -definitionen fr die Verwendung
		von System-Timern

	Funktionsweise:
		siehe timer.S

***********************************************************************************/

#ifndef TIMER_H_INCLUDED
#define TIMER_H_INCLUDED

#include "message.h"

#define MAX_TIMERS			8		// Anzahl der Timer-Slots
#define TIMER_STRUCT_SIZE	(MESSAGE_STRUCT_SIZE+4)	// Gre der TIMER-Struktur
#define SYSTEM_TIMER_TICKS	18432	// 14745600 MHz / 8 / 18432 = 100 Hz = jede 10 ms ein IRQ

#ifndef __ASSEMBLER__

typedef struct
{
	Message msg;			// Message des Timers
	uint16_t count;		// aktueller Zhlerwert des Timers, wird in jeder Timer-ISR runtergezhlt
	uint16_t maxval;	// Maximalwert, den der Timer haben kann
} __attribute__((packed)) __TIMER, *HTIMER;

extern volatile __TIMER _timers[MAX_TIMERS];	// Array der Timer-Eintrge
extern volatile __TIMER* volatile _timerindex;	// Index, wo sich der nchste freie Slot befinden


#ifdef __cplusplus
extern "C" {
#endif
/*******************************************************************
	Lsst jede cout * 10ms eine Message schicken
	 - wenn pfunc == NULL dann wird die Standard-Prozedur auferufen
	 - sonst pfunc
	Rckgabewerte:
		- Handle zum Timer bei Erfolg
		- NULL bei Misserfolg, z.B. wenn keine Timer mehr da sind
********************************************************************/
HTIMER SetTimer(uint16_t cout, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param = 0);

/*******************************************************************
	Lsst nur 1x nach cout * 10ms eine Message schicken
	 - wenn pfunc == NULL dann wird die Standard-Prozedur auferufen
	 - sonst pfunc
	Rckgabewerte:
		- Handle zum Timer bei Erfolg
		- NULL bei Misserfolg, z.B. wenn keine Timer mehr da sind
********************************************************************/
HTIMER SetAlarm(uint16_t count, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param  = 0);

/*******************************************************************
	Aktiviert den angegebenen Thread nach count * 10ms
	Rckgabewerte:
		- Handle zum Timer bei Erfolg
		- NULL bei Misserfolg, z.B. wenn keine Timer mehr da sind
********************************************************************/
HTIMER SetResumeAlarm(uint16_t, HTHREAD);

/********************************************************************
	Killt den angegebenen Timer
	Rckgabewerte:
		-  0 bei Erfolg
		- -1 bei Misserfolg
********************************************************************/
int16_t KillTimer(HTIMER);


/********************************************************************
	Gibt den count-Wert des Timers zurck
********************************************************************/
uint16_t GetTimerCount(HTIMER);

HTIMER SetStopwatch();

int16_t ResetTimer(HTIMER, uint16_t cout, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param = 0);
int16_t ResetAlarm(HTIMER, uint16_t count, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param  = 0);
int16_t ResetStopwatch(HTIMER);

#ifdef __cplusplus
}
#endif

/**********************************************************************
	Timer-Klasse fr einfache Verwendung von Timern
***********************************************************************/
class Timer
{
private:
	HTIMER ht;
public:
	inline Timer(): ht(NULL){}
	inline int16_t Set(uint16_t count, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param = 0)
	{
		if(ht)
			return ResetTimer(ht, count, pfunc, msg, param);

		if((ht = SetTimer(count, pfunc, msg, param)))
			return 0;
		return 1;
	}

	inline int16_t Alarm(uint16_t count, void(*pfunc)(int16_t, int16_t) = 0, int16_t msg = MSG_TIMER, int16_t param = 0)
	{
		if(ht)
			return ResetAlarm(ht, count, pfunc, msg, param);

		if((ht = SetAlarm(count, pfunc, msg, param)))
			return 0;
		return 1;
	}

	inline void Kill()
	{
		KillTimer(ht);
		ht = 0;
	}

	inline operator bool()
	{
		return ht;
	}

	inline operator uint16_t()
	{
		return GetTimerCount(ht);
	}
};

#endif // __ASSEMBLER__

#endif // TIMER_H_INCLUDED
