/***********************************************************************************
	
	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		fifo.h
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.
	
	Beschreibung: 
		Funktionsdeklarationen und -definitionen fr die Verwendung
		von FIFOs
	
***********************************************************************************/

#ifndef FIFO_H_INCLUDED
#define FIFO_H_INCLUDED

#include <stdlib.h>
#include <avr/interrupt.h>

#ifndef __ASSEMBLER__

#include <stdint.h>

typedef struct 
{
	uint8_t *buffer;
	size_t size;
	uint8_t *p_in;
	uint8_t *p_out;
} __attribute__((packed)) Fifo;


/**************************************************************
	Initialisiert den FIFO, buffer muss vorher reserviert werden
	Rckgabewerte:
		-  0: FIFO erfolgreich initialisiert
		- -1: fifo oder buffer oder size sind 0
 *************************************************************/
inline int8_t InitFifo(Fifo *fifo, uint8_t *buffer, size_t size)
{
	if((!fifo) || (!buffer) || (!size))
		return -1;
	fifo->buffer = buffer;
	fifo->size = size;
	fifo->p_in = fifo->p_out = buffer+size;
	return 0;
}

/*************************************************************
	Legt einen Wert in den Fifo
	 - Fr ISRs, Interrupts werden nicht deaktiviert
	 - Iniline um Pro-/Epilog klein zu halten 
 *************************************************************/
inline void PushFifoFast(Fifo *fifo, uint8_t c)
{
	*(--fifo->p_in) = c;
	if(fifo->p_in<=fifo->buffer)
		fifo->p_in = fifo->buffer+fifo->size;
	if(fifo->p_in==fifo->p_out)
	{
		fifo->p_out--;
		if(fifo->p_out<=fifo->buffer)
			fifo->p_out = fifo->buffer+fifo->size;
	}
}

/**************************************************************
	Holt einen vom Fifo
	 - Fr ISRs, Interrupts werden nicht deaktiviert
	 - Iniline um Pro-/Epilog klein zu halten 
	 - geht davon aus, dass im Fifo mind. ein Wert ist
 *************************************************************/
inline uint8_t PopFifoFast(Fifo *fifo)
{
	uint8_t c = *(--fifo->p_out);
	if(fifo->p_out<=fifo->buffer)
		fifo->p_out = fifo->buffer+fifo->size;
	return c;
}

/**************************************************************
	Gibt die gre des Fifos zurck
**************************************************************/
inline size_t SizeOfFifo(Fifo *fifo)
{
	return fifo->size;
}

/**************************************************************
	1 wenn leer, 0 wenn nicht
**************************************************************/
inline int8_t FifoEmpty(Fifo *fifo)
{
	uint8_t sreg = SREG;
	cli();
	int8_t c = fifo->p_in == fifo->p_out;
	SREG = sreg;
	return c;
}

#ifdef __cplusplus
extern "C" {
#endif
/***************************************************************
	Legt einen Wert in den Fifo
***************************************************************/
void PushFifo(Fifo *fifo, uint8_t c);	

/***************************************************************
	Wartet, bis mind. 1 Wert im Fifo ist und gibt diesen zurck
***************************************************************/
uint8_t PopFifo(Fifo* fifo);

/***************************************************************
	Gibt die Anzal der abgelegten Byts im Fifo zurck
***************************************************************/
size_t BytesInFifo(Fifo *fifo);
#ifdef __cplusplus
}
#endif

#endif // __ASSEMBLER__

#endif // FIFO_H_INCLUDED
