
/***********************************************************************************

	Projekt:	AVR-OS - Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		mutex.cpp
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.
	
	Beschreibung:
		Mutex-Funktionen zur Synchronisation von Prozessen

	Funktionsweise:
		Mutex verhindert, dass mehrere Threads gleichzeitig auf eine
		Ressource zugreifen. Dazu wird vor dem Zugriff die Mutex-Variable gesperrt
		und danach wieder freigegeben. Sollte whrenddessen ein anderer Thread
		versuchen, die Mutex-Variable zu sperren, wird dieser Thread bis zu
		ihrer Freigabe blockiert (passives Warten). Zustzlich kann man auf die
		Freigabe einer Mutex-Variable warten, ohne sie danach direkt zu sperren.

		Um Laufzeit und Speicherverbrauch mglichst gering zu halten, wurden keine
		Sicherheitmechanismen eingebaut, die Deaklock oder hnliches verhindern.
		Eine Mutex-Variable kann von JEDEM Thread zu jeder Zeit freigegeben werden.

***********************************************************************************/
#include <avr/interrupt.h>
#include <stdint.h>
#include "mutex.h"

void InitMutex(Mutex* m, int8_t state)
{
	int8_t sreg = DisableIRQ();
	m->state = state;
	m->wait_queue = NULL;
	RestoreIRQ(sreg);
}

void LockMutex(Mutex *m)
{
	int8_t sreg = DisableIRQ();

	while(m->state != MUTEX_STATE_UNLOCKED)
	{
		MutexWaitElement mwe;	// sich in die Liste eintrtagen
		mwe.next = m->wait_queue;
		mwe.hthread = GetCurrentThread();
		m->wait_queue = &mwe;
		SuspendOwnThread(); // Interrupts werden automatisch aktiviert
		sreg = DisableIRQ();
	}
	
	m->state = MUTEX_STATE_LOCKED; // Mutex sperren
	m->wait_queue = NULL;
	
	RestoreIRQ(sreg);
}

int8_t TryLockMutex(Mutex *m)
{
	int8_t sreg = DisableIRQ();
	int8_t ret = -1;

	if(m->state == MUTEX_STATE_UNLOCKED)
	{
		m->state = MUTEX_STATE_LOCKED;
		m->wait_queue = NULL;
		ret = 0;
	}

	RestoreIRQ(sreg);
	return ret;
}

void ReleaseMutex(Mutex* m)
{
	int8_t sreg = DisableIRQ();

	m->state = MUTEX_STATE_UNLOCKED;

	MutexWaitElement *wait_queue = m->wait_queue;
	m->wait_queue = NULL;

	while(wait_queue)	// Jedes Tlement durchgehen
	{
		ResumeThread(wait_queue->hthread);	// und den jeweiligen 
		wait_queue = wait_queue->next;		// Thread aktivieren
	}

	RestoreIRQ(sreg);
}

void WaitMutex(Mutex* m)
{
	int8_t sreg = DisableIRQ();

	if(m->state == MUTEX_STATE_LOCKED)
	{
		MutexWaitElement mwe;	// sich in die Liste eintrtagen
		mwe.next = m->wait_queue;
		mwe.hthread = GetCurrentThread();
		m->wait_queue = &mwe;
		SuspendOwnThread(); // Interrupts werden automatisch aktiviert
	}
	else
		RestoreIRQ(sreg);
}
