/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		os.h
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Funktionsdeklarationen und -definitionen fr den Betriebssystem-Kern

	Funktionsweise:
		siehe syscalls.S

***********************************************************************************/

#ifndef OS_H_INCLUDED
#define OS_H_INCLUDED

#include "fifo.h"
#include "message.h"


#define MAX_THREADS			8		// Number of thread slots
#define THREAD_STRUCT_SIZE	3

#define THREAD_STATE_IDLE				(1<<0)	// Slot nicht besetzt
#define THREAD_STATE_WAITING			(1<<1)	// im FIFO, wartet CPU-Zeit
#define THREAD_STATE_ACTIVE				(1<<2)	// wird im Moment ausgefhrt
#define THREAD_STATE_UNINTERRUPTIBLE	(1<<3)	// nicht durch andere Threads unterbrechbar, jedoch druch IRQs
#define THREAD_STATE_BLOCKED			(1<<4)	// inaktiv. wartet auf Aktivierung

#define TIMER_PRESCALER				2		// Prescaler fr Timer1 = 8
#define SCHEDULER_TIMER_TICKS		1843	// Anzahl der Takte, die ein Thread maximal hat = 1843*8 = 14744

#ifndef __ASSEMBLER__

#include <avr/interrupt.h>
#include <avr/pgmspace.h>

typedef struct
{
	uint8_t *sp; 	 // Stackpointer
	uint8_t state; // Zustand
}  __attribute__((packed)) THREAD, *HTHREAD;

#include "timer.h"
#include "mutex.h"

extern volatile HTHREAD _active_thread;
extern volatile THREAD _main_thread;		// Hauptthread

extern volatile HTHREAD _thread_fifo[MAX_THREADS];	// FIFO for waiting treads
extern volatile HTHREAD *volatile _in_index, *volatile _out_index;	// in and out pointer for _thread_fifo

#ifdef __cplusplus
extern "C" {
#endif
int16_t CreateThread(HTHREAD, char*, size_t, int16_t(*)(void*), void*, uint8_t = 0);
int16_t ResumeThread(HTHREAD, unsigned char = 0);
int16_t SuspendThread(HTHREAD);
int16_t SuspendOwnThread();
int16_t WaitThread(HTHREAD, uint8_t = THREAD_STATE_IDLE);
void SwitchToThread();
int16_t TerminateThread(HTHREAD, int16_t);
void ExitThread(int16_t);
#ifdef __cplusplus
}
#endif

inline HTHREAD GetCurrentThread()	// Gibt das Handle des eigenen Threads zurck
{
	return _active_thread;
}

inline uint8_t EnterCriticalSection()	// Macht eingenen Thread ununterbrechbar durch andere
{
	uint8_t state = _active_thread->state;
	_active_thread->state = THREAD_STATE_UNINTERRUPTIBLE;
	return state;
}

inline void LeaveCriticalSection()	// Macht eingenen Thread unterbrechbar durch andere
{
	_active_thread->state = THREAD_STATE_ACTIVE;
}

inline void RestoreCriticalSection(uint8_t state)
{
	_active_thread->state = state;
}

inline void Sleep(uint16_t t)	// wartet t * 10ms
{
	cli();
	SetResumeAlarm(t, GetCurrentThread());
	SuspendOwnThread(); // Interrupts werden automatisch aktiviert
}

inline int16_t GetExitCodeThread(THREAD *ht)
{
	return (int16_t)ht->sp;
}

inline int8_t DisableIRQ()
{
	int8_t sreg = SREG;
	cli();
	return sreg;
}

inline void EnableIRQ()
{
	sei();
}

inline void RestoreIRQ(char sreg)
{
	SREG = sreg;
}


#endif // __ASSEMBLER__

#endif // OS_H_INCLUDED
