/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		syscalls.S
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Implementiert den Betriebssystem-Kern

	Funktionsweise:
		Der Kern des Betriebsystem-Kerns ist der Scheduler. Dieser wird periodisch
		durch einen Timer aufgerufen und guckt, ob in der Thread-Warteschlange
		Handles zu Threads vorhanden sind, die auf ihre Ausfhrung warten. Sollte 
		dies der Fall sein und sollte der momentan aktive Thread unterbrechbar 
		sein (thread->state != THREAD_STATE_UNINTERRUPTIBLE) wird der momentan 
		aktive Thread ans Ende der Warteschlange gelegt und der nchste Thread
		aktiviert, indem der alte Stackpointer gespeichert und ein neuer geladen
		wird.
		
		Um neue Threads zu erzeugen, muss eine Funktion CreateThread()  aufrufen 
		den Stack-Speicher fr den Thread (mind. 37 Bytes,  64-256 Empfohlen) 
		und eine THREAD-Struktur, in der der Stackpointer und der Status 
		gespeichert wird, selbst bereitstellen. 
		
		Man kann Threads mittels SuspendThread() schlafen legen und mit 
		ResumeThread() wieder aufwecken. Mann kann sie auch jederzeit mit
		TerminateThread() beenden. 
		
		Muss ein Thread auf ein Ereignis durch Polling warten, so solle er 
		nach jeder Abfrage SwitchToThread() aufrufen, um die CPU anderen Threads
		zu geben, bis der Thread irgendwannmal wieder selbst an der Reihe ist.

***********************************************************************************/

#include <avr/io.h>
#include "os.h"
#include "timer.h"

/**************************  Scheduler ISR  ******************************/
.global TIMER1_COMPA_vect
.func TIMER1_COMPA_vect
TIMER1_COMPA_vect:

	push YL
	in YL, _SFR_IO_ADDR(SREG)
	push YL
	push YH

	lds YL, OCR1AL
	lds YH, OCR1AH
	subi YL, lo8(-SCHEDULER_TIMER_TICKS)
	sbci YH, hi8(-SCHEDULER_TIMER_TICKS)
	sts OCR1AH, YH
	sts OCR1AL, YL

	lds YL, _in_index
	lds YH, _out_index

	cp YL, YH
	brne do_other_tasks
	rjmp no_other_tasks

do_other_tasks:
	push XL

	lds YL, _active_thread
	lds YH, _active_thread+1

	clr XL
	cp YL, XL
	cpc YH, XL
	brne scheduler_thread_active

	push XH	// kein thread aktiv
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 24
	push 25

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))

	rjmp sched_act_pop_task

scheduler_thread_active:
	ldd  XL, Y+2

	cpi XL, THREAD_STATE_UNINTERRUPTIBLE
	brne do_task_switch
	rjmp no_task_switch

do_task_switch:
	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 24
	push 25

	ldi 16, THREAD_STATE_WAITING	// status speichern
	std Y+2, 16
	in XL, _SFR_IO_ADDR(SPL)
	in XH, _SFR_IO_ADDR(SPH)
	std Y+0, XL
	std Y+1, XH

	lds ZL, _in_index			// Z = _in_index
	lds ZH, _in_index+1

	st Z+, YL					// push(_active_thread)
	st Z+, YH

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi ZL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc ZH, 16
	brne no_in_overflow
	ldi ZL, lo8(_thread_fifo)
	ldi ZH, hi8(_thread_fifo)
no_in_overflow:
	sts _in_index, ZL
	sts _in_index+1, ZH

//--------- gespeichert ---------------

sched_act_pop_task:
	lds ZL, _out_index
	lds ZH, _out_index+1

	lds XL, _in_index
	lds XH, _in_index+1

pop_task:
	ld YL, Z+
	ld YH, Z+
	cpi ZL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc ZH, 16
	brne no_out_overflow
	ldi ZL, lo8(_thread_fifo)
	ldi ZH, hi8(_thread_fifo)
no_out_overflow:

	cp ZL, XL
	cpc ZH, XH
	breq sched_pop_regs

	ldd 17, Y+2
	cpi 17, THREAD_STATE_WAITING
	brne pop_task

	ldi 16, THREAD_STATE_ACTIVE
	std Y+2, 16

	sts _active_thread, YL
	sts _active_thread+1, YH

	ldd XL, Y+0
	ldd XH, Y+1
	out _SFR_IO_ADDR(SPL), XL
	out _SFR_IO_ADDR(SPH), XH

sched_pop_regs:
	sts _out_index, ZL
	sts _out_index+1, ZH

	pop 25
	pop 24
	pop 23
	pop 22
	pop 21
	pop 20
	pop 19
	pop 18
	pop 17
	pop 16
	pop 15
	pop 14
	pop 13
	pop 12
	pop 11
	pop 10
	pop 9
	pop 8
	pop 7
	pop 6
	pop 5
	pop 4
	pop 3
	pop 2
	pop 1
	pop 0
	pop ZH
	pop ZL
	pop XH
no_task_switch:
	pop XL
no_other_tasks:
	pop YH
	pop YL
	out _SFR_IO_ADDR(SREG), YL
	pop YL
	reti
.endfunc

/******************************** SwitchToThread() ******************************/
.global SwitchToThread
.func SwitchToThread
SwitchToThread:
	cli
	push YL
	in YL, _SFR_IO_ADDR(SREG)
	push YL
	push YH

.global SwitchToThread_inside
SwitchToThread_inside:

	lds YL, _in_index
	lds YH, _out_index

	cp YL, YH
	brne switch_do_other_tasks
	rjmp switch_no_other_tasks

switch_do_other_tasks:
	push XL

	lds YL, _active_thread
	lds YH, _active_thread+1

	clr XL
	cp YL, XL
	cpc YH, XL
	brne switch_thread_active

	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 24
	push 25

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))

	rjmp switch_act_pop_task

switch_thread_active:
	ldd  XL, Y+2

	cpi XL, THREAD_STATE_UNINTERRUPTIBLE
	brne switch_do_task_switch
	rjmp switch_no_task_switch

 switch_do_task_switch:
	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 24
	push 25

	ldi 16, THREAD_STATE_WAITING	// status speichern
	std Y+2, 16
	in XL, _SFR_IO_ADDR(SPL)
	in XH, _SFR_IO_ADDR(SPH)
	std Y+0, XL
	std Y+1, XH

	lds ZL, _in_index			// Z = _in_index
	lds ZH, _in_index+1

	st Z+, YL					// push(_active_thread)
	st Z+, YH

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi ZL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc ZH, 16
	brne switch_no_in_overflow
	ldi ZL, lo8(_thread_fifo)
	ldi ZH, hi8(_thread_fifo)
switch_no_in_overflow:
	sts _in_index, ZL
	sts _in_index+1, ZH

//--------- gespeichert ---------------

switch_act_pop_task:
	lds ZL, _out_index
	lds ZH, _out_index+1

	lds XL, _in_index
	lds XH, _in_index+1

switch_pop_task:
	ld YL, Z+
	ld YH, Z+
	cpi ZL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc ZH, 16
	brne switch_no_out_overflow
	ldi ZL, lo8(_thread_fifo)
	ldi ZH, hi8(_thread_fifo)
switch_no_out_overflow:

	ldd 17, Y+2
	cpi 17, THREAD_STATE_WAITING
	breq switch_got_task

	cp ZL, XL
	cpc ZH, XH
	breq switch_pop_regs

	rjmp switch_pop_task

switch_got_task:
	ldi 16, THREAD_STATE_ACTIVE
	std Y+2, 16

	sts _active_thread, YL
	sts _active_thread+1, YH

	ldd XL, Y+0
	ldd XH, Y+1
	out _SFR_IO_ADDR(SPL), XL
	out _SFR_IO_ADDR(SPH), XH

switch_pop_regs:
	sts _out_index, ZL
	sts _out_index+1, ZH

#if defined (__AVR_ATmega644__)
	lds 16, TCNT1L		// OCR1A = TCNT1 + TIMER_TICKS
	lds 17, TCNT1H
#else
	in 16, _SFR_IO_ADDR(TCNT1L)		// OCR1A = TCNT1 + TIMER_TICKS
	in 17, _SFR_IO_ADDR(TCNT1H)
#endif

	subi 16, lo8(-SCHEDULER_TIMER_TICKS)
	sbci 17, hi8(-SCHEDULER_TIMER_TICKS)
	sts OCR1AH, 17
	sts OCR1AL, 16

	pop 25
	pop 24
	pop 23
	pop 22
	pop 21
	pop 20
	pop 19
	pop 18
	pop 17
	pop 16
	pop 15
	pop 14
	pop 13
	pop 12
	pop 11
	pop 10
	pop 9
	pop 8
	pop 7
	pop 6
	pop 5
	pop 4
	pop 3
	pop 2
	pop 1
	pop 0
	pop ZH
	pop ZL
	pop XH
switch_no_task_switch:
	pop XL
switch_no_other_tasks:
	pop YH
	pop YL
	out _SFR_IO_ADDR(SREG), YL
	pop YL
	reti
.endfunc

/************************* System-Timer ISR ********************************/
.global TIMER1_COMPB_vect
.func TIMER1_COMPB_vect
TIMER1_COMPB_vect:
	push YL
	in YL, _SFR_IO_ADDR(SREG)
	push YL
	push YH
	push ZL
	push ZH
	push 16
	push 17
	push 24
	push 25
	push 18
	push 19
	push 1

	clr 1
	clr 19

	lds YL, OCR1BL
	lds YH, OCR1BH
	subi YL, lo8(-SYSTEM_TIMER_TICKS)
	sbci YH, hi8(-SYSTEM_TIMER_TICKS)
	sts OCR1BH, YH
	sts OCR1BL, YL

	ldi ZL, lo8(_timers)
	ldi ZH, hi8(_timers)
	ldi 18, hi8(_timers+MAX_TIMERS*TIMER_STRUCT_SIZE)

timer_isr_next:
	ldd 24, Z+MESSAGE_STRUCT_SIZE+0
	ldd 25, Z+MESSAGE_STRUCT_SIZE+1
	cp 24, 1
	cpc 25, 1
	brne timer_isr_timer
	rjmp timer_isr_no_timer

timer_isr_timer:
	sbiw 24, 1
	cp 24, 1
	cpc 25, 1
	breq timer_isr_event
	rjmp timer_isr_no_event
timer_isr_event:

	ldd 24, Z+0
	ldd 25, Z+1
	cp 24, 1
	cpc 25, 1
	brne timer_isr_message	// if(pfunc) SendMessage() else ResumeThread()

	ldd YL, Z+2
	ldd YH, Z+3
	cp YL, 1
	cpc YH, 1
	brne timer_isr_resume_alarm
	rjmp timer_isr_no_event

timer_isr_resume_alarm:
	ldd 17, Y+2
	cpi 17, THREAD_STATE_BLOCKED
	brne timer_isr_thread_wrong		// if(hthread->state==THREAD_STATE_BLOCKED) ResumeThread()

	ldi 17, THREAD_STATE_WAITING
	std Y+2, 17

	movw 24, YL	// 24:25 waren NULL, jetzt HTHREAD

	lds YL, _out_index			// Y = _out_index	-> an den _Anfang_ des FIFOs anfgen!!!
	lds YH, _out_index+1

	sbiw YL, 2
	ldi 17, hi8(_thread_fifo)
	cpi YL, lo8(_thread_fifo)
	cpc YH, 17
	brcc timer_isr_resume_no_out_undewlow
	ldi YL, lo8(_thread_fifo+(MAX_THREADS*2)-2)
	ldi YH, hi8(_thread_fifo+(MAX_THREADS*2)-2)
timer_isr_resume_no_out_undewlow:
	sts _out_index, YL
	sts _out_index+1, YH

	st Y+, 24					// push(HTHREAD)
	st Y+, 25

	inc 19		// wenn am ende r19 != 0 dann SwitchToThread()

timer_isr_thread_wrong:
	ldd 24, Z+MESSAGE_STRUCT_SIZE+2
	ldd 25, Z+MESSAGE_STRUCT_SIZE+3

	cp 24, 1
	cpc 25, 1
	brne timer_isr_no_event

	lds 16, _timerindex
	lds 17, _timerindex+1
	cp ZL, 16
	cpc ZH, 17
	brcc timer_isr_no_event	// if(_timerindex > ptimer) _timerindex = ptimer
	sts _timerindex, ZL
	sts _timerindex+1, ZH

	rjmp timer_isr_no_event

timer_isr_message:
	// SendFastMessage
	lds YL, message_out_index
	lds YH, message_out_index+1
	sbiw YL, MESSAGE_STRUCT_SIZE

	ldi 16, hi8(message_fifo)
	cpi YL, lo8(message_fifo)
	cpc YH, 16
	brcc timer_isr_no_undeflow
	ldi YL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE-MESSAGE_STRUCT_SIZE)
	ldi YH, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE-MESSAGE_STRUCT_SIZE)
timer_isr_no_undeflow:
	sts message_out_index, YL
	sts message_out_index+1, YH

	std Y+0, 24		// r24:r25 schon vorher mit pfunc geladen
	std Y+1, 25
	ldd 17, Z+2
	std Y+2, 17
	ldd 17, Z+3
	std Y+3, 17
	ldd 17, Z+4
	std Y+4, 17
	ldd 17, Z+5
	std Y+5, 17

	ldd 24, Z+MESSAGE_STRUCT_SIZE+2
	ldd 25, Z+MESSAGE_STRUCT_SIZE+3

	cp 24, 1
	cpc 25, 1
	brne timer_isr_no_event

	lds 16, _timerindex
	lds 17, _timerindex+1
	cp ZL, 16
	cpc ZH, 17
brcc timer_isr_no_event	// if(_timerindex > ptimer) _timerindex = ptimer
	sts _timerindex, ZL
	sts _timerindex+1, ZH

timer_isr_no_event:
	std Z+MESSAGE_STRUCT_SIZE+0, 24
	std Z+MESSAGE_STRUCT_SIZE+1, 25

timer_isr_no_timer:
	adiw ZL, TIMER_STRUCT_SIZE
	cpi ZL, lo8(_timers+MAX_TIMERS*TIMER_STRUCT_SIZE)
	cpc ZH, 18
	brcc timer_isr_done
	rjmp timer_isr_next
timer_isr_done:

	tst 19
	brne timer_isr_switch_return	// if any ThreadAlarms then return SwitchToThread()

	pop 1
	pop 19
	pop 18
	pop 25
	pop 24
	pop 17
	pop 16
	pop ZH
	pop ZL
	pop YH
	pop YL
	out _SFR_IO_ADDR(SREG), YL
	pop YL
	reti

timer_isr_switch_return:
	pop 1
	pop 19
	pop 18
	pop 25
	pop 24
	pop 17
	pop 16
	pop ZH
	pop ZL
	jmp SwitchToThread_inside
.endfunc

/************************** CreateThread ****************************/
.global CreateThread
.func CreateThread		// 24:25: HTHREAD, 22:23: char*stack, 20:21 size,
					//18:19 void(*)(), 16:17: void*, 14: unsigned char mode
CreateThread:
	cp 18, 1					// prfen ob berhaupt eine Funktion angegeben wurde
	cpc 19, 1
	breq create_thread_failed

	movw ZL, 24 // Z = HTHREAD

	add 22, 20	// 22:23 = sp+stacksize
	adc 23, 21

	mov 20, 14 // 20 = mode

	std Z+0, 22	// ht->sp = sp
	std Z+1, 23

	movw 24, 18	// 24:25 = func

	in 0, _SFR_IO_ADDR(SREG)
	cli

	cpi 20, THREAD_STATE_ACTIVE
	breq create_active
	rjmp create_not_active

create_thread_failed:
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret

create_active:

	// prfen, ob eigener Thread unterbrechbar
	movw 18, ZL	// neuen handle zwischenspeichern

	lds ZL, _active_thread
	lds ZH, _active_thread+1

	ldd 21, Z+2
	cpi 21, THREAD_STATE_UNINTERRUPTIBLE
	breq dont_save_own_thread
	rjmp save_own_thread

dont_save_own_thread:	// neuen Thread an den Anfang des FIFOs hinzufgen
	movw ZL, 18

	ldd XL, Z+0			// Z = new HTHREAD
	ldd XH, Z+1			// X = SP des neuen Threads

	ldi 18, lo8(gs(_thread_end_proc))
	ldi 19, hi8(gs(_thread_end_proc))

	st -X, 18
	st -X, 19

	st -X, 24	// push(PC)
	st -X, 25

	st -X, YL	//
	st -X, 0	// SREG
	st -X, YH	// 0
	st -X, ZL	// 1
	st -X, ZH	// ZL
	st -X, ZL	// ZH
	st -X, ZH	// 18
	st -X, 0	// 2
	st -X, 1	// 3
	st -X, 2	// 4
	st -X, 3	// 5
	st -X, 4	// 6
	st -X, 5	// 7
	st -X, 6	// 8
	st -X, 7	// 9
	st -X, 8	// 10
	st -X, 9	// 11
	st -X, 10	// 12
	st -X, 11	// 13
	st -X, 12	// 14
	st -X, 13	// 15
	st -X, 14	// 19
	st -X, 15	// 20
	st -X, 16	// 21
	st -X, 17	// 22
	st -X, 18	// 23
	st -X, 19	// r24
	st -X, 20	// r25
	st -X, 21	// 26
	st -X, 22	// 27
	st -X, 23	// YL
	st -X, 16	// YH
	st -X, 17
	sbiw XL, 1

	std Z+0, XL
	std Z+1, XH

	ldi 18, THREAD_STATE_WAITING
	std Z+2, 18

	lds XL, _out_index			// X = _out_index	-> an den _Anfang_ des FIFOs anfgen!!!
	lds XH, _out_index+1

	sbiw XL, 2
	ldi 18, hi8(_thread_fifo)
	cpi XL, lo8(_thread_fifo)
	cpc XH, 18
	brcc create_no_out_undewlow
	ldi XL, lo8(_thread_fifo+(MAX_THREADS*2)-2)
	ldi XH, hi8(_thread_fifo+(MAX_THREADS*2)-2)
create_no_out_undewlow:
	sts _out_index, XL
	sts _out_index+1, XH

	st X+, ZL					// push(neuer Thread)
	st X+, ZH

	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret

save_own_thread:
	// eigenen thread speichern
	push YL
	push 0		// r0 = SREG
	push YH
	push XL
	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 1		// return 0
	push 1

	// 18:19 = neuer HTHREAD
	// Z = _active_thread

	movw YL, ZL
	movw ZL, 18

	in XL, _SFR_IO_ADDR(SPL)
	in XH, _SFR_IO_ADDR(SPH)

	std Y+0, XL					// eigenen stackpointer speichern
	std Y+1, XH
	ldi 16, THREAD_STATE_WAITING
	std Y+2, 16					// eigenen state speichern

	lds XL, _in_index			// X = _in_index
	lds XH, _in_index+1

	st X+, YL					// push(_active_thread)
	st X+, YH

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi XL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc XH, 16
	brne create_no_in_overflow
	ldi XL, lo8(_thread_fifo)
	ldi XH, hi8(_thread_fifo)
create_no_in_overflow:
	sts _in_index, XL
	sts _in_index+1, XH

	sts _active_thread, ZL
	sts _active_thread+1, ZH

	std Z+2, 20			// Z = neu HTHREAD, 20 = THREAD_STATE_ACTIVE

	ldd XL, Z+0			// X = SP des neuen Threads
	ldd XH, Z+1


	ldi 18, lo8(gs(_thread_end_proc))
	ldi 19, hi8(gs(_thread_end_proc))
	//lsr 19
	//ror 18

	st -X, 18
	st -X, 19
	sbiw XL, 1

	out _SFR_IO_ADDR(SPL), XL
	out _SFR_IO_ADDR(SPH), XH

	movw ZL, 24
	movw 24, 16

#if defined (__AVR_ATmega644__)
	lds 16, TCNT1L		// OCR1A = TCNT1 + TIMER_TICKS
	lds 17, TCNT1H
#else
	in 16, _SFR_IO_ADDR(TCNT1L)		// OCR1A = TCNT1 + TIMER_TICKS
	in 17, _SFR_IO_ADDR(TCNT1H)
#endif

	subi 16, lo8(-SCHEDULER_TIMER_TICKS)
	sbci 17, hi8(-SCHEDULER_TIMER_TICKS)

	sts OCR1AH, 17
	sts OCR1AL, 16
	
	ldi 16, (1<<OCF1A)
#if defined (__AVR_ATmega644__)
	out _SFR_IO_ADDR(TIFR1), 16
#else
	out _SFR_IO_ADDR(TIFR), 16
#endif
	
	sei
	ijmp

create_not_active:

	ldd XL, Z+0			// Z = new HTHREAD
	ldd XH, Z+1			// X = SP des neuen Threads

	ldi 18, lo8(gs(_thread_end_proc))
	ldi 19, hi8(gs(_thread_end_proc))
	//lsr 19
	//ror 18

	st -X, 18
	st -X, 19

	st -X, 24	// push(PC)
	st -X, 25

	st -X, YL	//
	st -X, 0	// SREG
	st -X, YH	// 0
	st -X, ZL	// 1
	st -X, ZH	// ZL
	st -X, ZL	// ZH
	st -X, ZH	// 18
	st -X, 0	// 2
	st -X, 1	// 3
	st -X, 2	// 4
	st -X, 3	// 5
	st -X, 4	// 6
	st -X, 5	// 7
	st -X, 6	// 8
	st -X, 7	// 9
	st -X, 8	// 10
	st -X, 9	// 11
	st -X, 10	// 12
	st -X, 11	// 13
	st -X, 12	// 14
	st -X, 13	// 15
	st -X, 14	// 19
	st -X, 15	// 20
	st -X, 16	// 21
	st -X, 17	// 22
	st -X, 18	// 23
	st -X, 19	// r24
	st -X, 20	// r25
	st -X, 21	// 26
	st -X, 22	// 27
	st -X, 23	// YL
	st -X, 16	// YH
	st -X, 17
	sbiw XL, 1

	std Z+0, XL
	std Z+1, XH

	cpi 20, THREAD_STATE_BLOCKED
	brne create_start_thread

	std Z+2, 20
	rjmp create_dont_start_thread

create_start_thread:
	ldi 18, THREAD_STATE_WAITING
	std Z+2, 18

	lds XL, _in_index			// X = _in_index
	lds XH, _in_index+1

	st X+, ZL					// push(neuer Thread)
	st X+, ZH

	ldi 18, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi XL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc XH, 18
	brne create_no_in_overflow_2
	ldi XL, lo8(_thread_fifo)
	ldi XH, hi8(_thread_fifo)
create_no_in_overflow_2:
	sts _in_index, XL
	sts _in_index+1, XH

create_dont_start_thread:
	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret
.endfunc

/****************************+** ResumeThread **************************************/
.global ResumeThread		// r24:r25 HTHREAD, r22 mode
.func ResumeThread
ResumeThread:
	movw ZL, 24

	in 0, _SFR_IO_ADDR(SREG)
	cli

	ldd 18, Z+2
	cpi 18, THREAD_STATE_BLOCKED
	brne resume_thread_in_wrong_state

	cpi 22, THREAD_STATE_ACTIVE
	breq resume_active

	ldi 18, THREAD_STATE_WAITING
	std Z+2, 18

	lds XL, _in_index			// X = _in_index
	lds XH, _in_index+1

	st X+, ZL					// push(neuer Thread)
	st X+, ZH

	ldi 18, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi XL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc XH, 18
	brne resume_no_in_overflow
	ldi XL, lo8(_thread_fifo)
	ldi XH, hi8(_thread_fifo)
resume_no_in_overflow:
	sts _in_index, XL
	sts _in_index+1, XH

	out _SFR_IO_ADDR(SREG), 0

	clr 24
	clr 25
	ret

resume_thread_in_wrong_state:
	out _SFR_IO_ADDR(SREG), 0
	ldi 24, lo8(-2)
	ldi 25, hi8(-2)
	ret

resume_active:
	// prfen, ob eigener Thread unterbrechbar
	movw 18, ZL	// handle zwischenspeichern

	lds ZL, _active_thread
	lds ZH, _active_thread+1

	ldd 21, Z+2
	cpi 21, THREAD_STATE_UNINTERRUPTIBLE
	breq resume_dont_save_own_thread
	rjmp resume_save_own_thread

resume_dont_save_own_thread:	// Thread an den Anfang des FIFOs hinzufgen
	movw ZL, 18

	ldi 18, THREAD_STATE_WAITING
	std Z+2, 18

	lds XL, _out_index			// X = _out_index	-> an den _Anfang_ des FIFOs anfgen!!!
	lds XH, _out_index+1

	sbiw XL, 2
	ldi 18, hi8(_thread_fifo)
	cpi XL, lo8(_thread_fifo)
	cpc XH, 18
	brcc resume_no_out_undewlow
	ldi XL, lo8(_thread_fifo+(MAX_THREADS*2)-2)
	ldi XH, hi8(_thread_fifo+(MAX_THREADS*2)-2)
resume_no_out_undewlow:
	sts _out_index, XL
	sts _out_index+1, XH

	st X+, ZL					// push(neuer Thread)
	st X+, ZH

	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret

resume_save_own_thread:
	// eigenen thread speichern
	push YL
	push 0		// r0 = SREG
	push YH
	push XL
	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 1		// return 0
	push 1

	// 18:19 = HTHREAD
	// Z = _active_thread

	movw YL, ZL
	movw ZL, 18

	in XL, _SFR_IO_ADDR(SPL)
	in XH, _SFR_IO_ADDR(SPH)

	std Y+0, XL					// eigenen stackpointer speichern
	std Y+1, XH
	ldi 16, THREAD_STATE_WAITING
	std Y+2, 16					// eigenen state speichern

	lds XL, _in_index			// X = _in_index
	lds XH, _in_index+1

	st X+, YL					// push(_active_thread)
	st X+, YH

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))
	cpi XL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc XH, 16
	brne resume_no_in_overflow_2
	ldi XL, lo8(_thread_fifo)
	ldi XH, hi8(_thread_fifo)
resume_no_in_overflow_2:
	sts _in_index, XL
	sts _in_index+1, XH

	sts _active_thread, ZL
	sts _active_thread+1, ZH

	std Z+2, 22			// Z = neu HTHREAD, 22 = THREAD_STATE_ACTIVE

	ldd XL, Z+0			// X = SP des neuen Threads
	ldd XH, Z+1

	out _SFR_IO_ADDR(SPL), XL
	out _SFR_IO_ADDR(SPH), XH

#if defined (__AVR_ATmega644__)
	lds 16, TCNT1L
	lds 17, TCNT1H
#else
	in 16, _SFR_IO_ADDR(TCNT1L)		// OCR1A = TCNT1 + TIMER_TICKS
	in 17, _SFR_IO_ADDR(TCNT1H)
#endif

	subi 16, lo8(-SCHEDULER_TIMER_TICKS)
	sbci 17, hi8(-SCHEDULER_TIMER_TICKS)
	sts OCR1AL, 16
	sts OCR1AH, 17

	ldi 16, (1<<OCF1A)
#if defined (__AVR_ATmega644__)
	out _SFR_IO_ADDR(TIFR1), 16
#else
	out _SFR_IO_ADDR(TIFR), 16
#endif

	pop 25
	pop 24
	pop 23
	pop 22
	pop 21
	pop 20
	pop 19
	pop 18
	pop 17
	pop 16
	pop 15
	pop 14
	pop 13
	pop 12
	pop 11
	pop 10
	pop 9
	pop 8
	pop 7
	pop 6
	pop 5
	pop 4
	pop 3
	pop 2
	pop 1
	pop 0
	pop ZH
	pop ZL
	pop XH
	pop XL
	pop YH
	pop YL
	out _SFR_IO_ADDR(SREG), YL
	pop YL
	reti
.endfunc


/******************************** SuspendThread() *******************************/
.global SuspendThread		// r24:r25 = HTHREAD
.func SuspendThread
SuspendThread:
	lds ZL, _active_thread
	lds ZH, _active_thread+1
	cp ZL, 24
	cpc ZH, 25
	breq SuspendOwnThread_inside

	movw ZL, 24		// Z = hThread

	ldi 18, THREAD_STATE_BLOCKED

	in 0, _SFR_IO_ADDR(SREG)
	cli

	ldd 19, Z+2		// prfen, ob hThread->state == THREAD_STATE_WAITING
	cpi 19, THREAD_STATE_WAITING
	brne suspend_wrong_state
	out _SFR_IO_ADDR(SREG), 0	// Interrupt erst nach nchstem Befehl aktiv
	std Z+2, 18

	clr 24
	clr 25
	ret

suspend_wrong_state:
	out _SFR_IO_ADDR(SREG), 0
	ldi 24, lo8(-2)
	ldi 25, hi8(-2)
	ret
.endfunc

/******************************* SuspendOwnThread() *****************************/
.global SuspendOwnThread
.func SuspendOwnThread
SuspendOwnThread:
	lds ZL, _active_thread
	lds ZH, _active_thread+1
SuspendOwnThread_inside:		// fr SuspendThread
	ldd 18, Z+2
	cpi 18, THREAD_STATE_UNINTERRUPTIBLE
	brne suspendown_write_state
	
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret
	
suspendown_write_state:
	cli

	push YL
	in YL, _SFR_IO_ADDR(SREG)
	push YL
	push YH
	push XL
	push XH
	push ZL
	push ZH
	push 0
	push 1
	push 2
	push 3
	push 4
	push 5
	push 6
	push 7
	push 8
	push 9
	push 10
	push 11
	push 12
	push 13
	push 14
	push 15
	push 16
	push 17
	push 18
	push 19
	push 20
	push 21
	push 22
	push 23
	push 1		// return 0
	push 1

	ldi 18, THREAD_STATE_BLOCKED

	in XL, _SFR_IO_ADDR(SPL)
	in XH, _SFR_IO_ADDR(SPH)

	std Z+0, XL	// Stack speichern
	std Z+1, XH
	std Z+2, 18
	sts _active_thread,1		// kein Thread aktiv
	sei
	sts _active_thread+1,1

	jmp _thread_end_proc_inside
.endfunc

/********************** int WaitThread(HTHREAD, attr) **************************/
.global WaitThread
.func WaitThread
WaitThread:
	movw ZL, 24

wait_thread_loop:
	ldd 18, Z+2
	and 18, 22
	brne wait_ready
	call SwitchToThread
	rjmp wait_thread_loop

wait_ready:
	clr 24
	clr 25
	ret
.endfunc


/******************************* TerminateThread() *****************************/
.global TerminateThread	// int TerminateThread(24:25 HTHREAD, 22:23: unsigned short)
.func TerminateThread
TerminateThread:
	lds ZL, _active_thread
	lds ZH, _active_thread+1
	cp ZL, 24
	cpc ZH, 25
	brne kill_other_thread
	movw 24, 22
	jmp _thread_end_proc	// wenn eigener Thread dann sich selbst killen

kill_other_thread:

	in 0, _SFR_IO_ADDR(SREG)
	cli

	std Z+0, 22	// save Exit-Code
	std Z+1, 23

	movw ZL, 24
	ldi 18, THREAD_STATE_IDLE
	std Z+2, 18


	out _SFR_IO_ADDR(SREG), 0

	clr 24
	clr 25
	ret
.endfunc

/******************************* ExitThread() *****************************/
.global ExitThread
.func ExitThread
ExitThread:
	jmp _thread_end_proc
.endfunc

/******************************** _thread_end_proc ******************************/
.section .fini8
.global _thread_end_proc
.func _thread_end_proc
_thread_end_proc:
	lds ZL, _active_thread		// eigenen Thread als idle speichern
	lds ZH, _active_thread+1

	ldi 18, THREAD_STATE_IDLE
	cli
	std Z+2, 18					// r18: state des aktiven Threads

	std Z+0, 24					// Exit-Code speichern
	std Z+1, 25

	sts _active_thread,1		// kein Thread aktiv
	sts _active_thread+1,1

_thread_end_proc_inside:
	cli
	lds XL, _in_index
	lds ZL, _out_index
	cp XL, ZL
	brne task_in_fifo
	sei
	nop
	nop
	nop
	rjmp _thread_end_proc_inside
task_in_fifo:
	lds XH, _in_index+1		// X = _in_index
	lds ZH, _out_index+1	// Z = _out_index

	ldi 16, hi8(_thread_fifo+(MAX_THREADS*2))
_thread_end_pop_task:
	ld YL, Z+
	ld YH, Z+
	cpi ZL, lo8(_thread_fifo+(MAX_THREADS*2))
	cpc ZH, 16
	brne _thread_end_no_out_overflow
	ldi ZL, lo8(_thread_fifo)
	ldi ZH, hi8(_thread_fifo)
_thread_end_no_out_overflow:

	ldd 17, Y+2
	cpi 17, THREAD_STATE_WAITING
	breq _thread_end_got_task

	cp XL, ZL
	cpc XH, ZH
	brne _thread_end_pop_task

	sts _out_index, ZL			// if(fifo.empty())
	sts _out_index+1, ZH
	sei
	rjmp _thread_end_proc_inside
	
_thread_end_got_task:
	sts _out_index, ZL
	sts _out_index+1, ZH

	ldi 16, THREAD_STATE_ACTIVE
	std Y+2, 16

	sts _active_thread, YL
	sts _active_thread+1, YH

	ldd XL, Y+0
	ldd XH, Y+1
	out _SFR_IO_ADDR(SPL), XL
	out _SFR_IO_ADDR(SPH), XH

#if defined (__AVR_ATmega644__)
	lds 16, TCNT1L
	lds 17, TCNT1H
#else
	in 16, _SFR_IO_ADDR(TCNT1L)		// OCR1A = TCNT1 + TIMER_TICKS
	in 17, _SFR_IO_ADDR(TCNT1H)
#endif

	subi 16, lo8(-SCHEDULER_TIMER_TICKS)
	sbci 17, hi8(-SCHEDULER_TIMER_TICKS)
	sts OCR1AL, 16
	sts OCR1AH, 17

	ldi 16, (1<<OCF1A)
#if defined (__AVR_ATmega644__)
	out _SFR_IO_ADDR(TIFR1), 16
#else
	out _SFR_IO_ADDR(TIFR), 16
#endif

	pop 25
	pop 24
	pop 23
	pop 22
	pop 21
	pop 20
	pop 19
	pop 18
	pop 17
	pop 16
	pop 15
	pop 14
	pop 13
	pop 12
	pop 11
	pop 10
	pop 9
	pop 8
	pop 7
	pop 6
	pop 5
	pop 4
	pop 3
	pop 2
	pop 1
	pop 0
	pop ZH
	pop ZL
	pop XH
	pop XL
	pop YH
	pop YL
	out _SFR_IO_ADDR(SREG), YL
	pop YL
	reti
.endfunc

.section .init8
.global _OS_STARTUP_CODE
.func _OS_STARTUP_CODE
_OS_STARTUP_CODE:
	ldi 16, THREAD_STATE_ACTIVE		// _main_thread.state = active
	sts _main_thread+2, 16

	ldi ZL, lo8(_main_thread)
	ldi ZH, hi8(_main_thread)
	sts _active_thread, ZL
	sts _active_thread+1, ZH

	ldi 16, lo8(SCHEDULER_TIMER_TICKS)	// Scheduler-Timer einstellen
	ldi 17, hi8(SCHEDULER_TIMER_TICKS)
	sts OCR1AH, 17
	sts OCR1AL, 16

	ldi 16, lo8(SYSTEM_TIMER_TICKS)		// System-Timer einstellen
	ldi 17, hi8(SYSTEM_TIMER_TICKS)
	sts OCR1BH, 17
	sts OCR1BL, 16

	ldi 16, TIMER_PRESCALER
#if defined (__AVR_ATmega644__)
	sts TCCR1B, 16
#else
	out _SFR_IO_ADDR(TCCR1B), 16
#endif

	ldi 16, (1<<OCIE1A) | (1<<OCIE1B)
#if defined (__AVR_ATmega644__)
	sts TIMSK1, 16
#else
	out _SFR_IO_ADDR(TIMSK), 16
#endif

	sei
.endfunc
