/***********************************************************************************
	
	Projekt:	AVR-OS - Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		main.c
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.
	
	
	Beschreibung: 
		Demonstriert das die Fhingkeiten des Betriebssystems AVR-OS
	
	Funktionsweise:
		Zuerst werden der UART und sein FIFO initialisiert. Denn werden 2 Threads
		gestartet - der eine Thread sendet per UART empfangene Zeichen wieder 
		zurck, der andere lsst ein 'X' in der zweiten Zeile des Terminals nach 
		rechts wandern und wartet nach jedem Schritt 110ms.
		
		In der main wird noch eine neue Standard-Prozedur gesetzt und ein Timer
		gestartet. Dieser sendet jede 170ms eine Nachricht, die ein weiteres 'X'
		in der dritten Zeile wandern lsst.  Damit die ankommenden Nachrichten 
		auch verarbeitet werden, holt sie das Hauptprogramm in einer Endlosschleife 
		ab und fhrt sie aus.
		
		Damit sich die Threads nicht gegenseitig stren, wird der Zugriff auf
		den UART mittels Mutex geregelt.

***********************************************************************************/

#include <avr/io.h>
#include <stdlib.h>
#include <avr-os/os.h>

#define ESC	27		// Escape Character

#define BAUD(n)	(F_CPU/(n)/16UL-1)

unsigned char uart_fifo_buffer[16];
Fifo uart_fifo;

/***********  UART-Funktionen  ********************************/

void InitUart(unsigned int baud)// Initialisiert den UART
{	
	UBRRH = baud>>8;
	UBRRL = baud;
	UCSRC = (1<<URSEL)|(3<<UCSZ0);
	UCSRB = (1<<TXEN)| (1<<RXEN) | (1<<RXCIE);
}

void UartPutChar(char c)		// Sendet ein Zeichen ber den UART
{
	while(!(UCSRA & (1<<UDRE)))
		SwitchToThread();
	
	UDR = c;
}

void UartPutStr(char *str)		// Sendet einen String ber den UART
{
	while(*str)
		UartPutChar(*str++);
}

char UartGetChar()				// Wartet auf ein Zeichen vom UART und gibd dieses zurck
{
	return PopFifo(&uart_fifo);
}

void GotoXY(int x, int y)		// Setzt den Cursor im Terminal auf Position x:y
{
	char szBuff[5];
	
	UartPutChar(ESC);
	UartPutChar('[');
	
	itoa(y, szBuff, 10);
	UartPutStr(szBuff);
	
	UartPutChar(';');
	
	itoa(x, szBuff, 10);
	UartPutStr(szBuff);
	
	UartPutChar('H');
}

void SaveCursor()	// Cursorposition mittels VT100-Kommando <ESC>[s speichern
{
	UartPutChar(ESC);
	UartPutChar('[');
	UartPutChar('s');
}

void RestoreCursor() // alte Cursorposition mittels VT100-Kommando <ESC>[u laden
{
	UartPutChar(ESC);
	UartPutChar('[');
	UartPutChar('u');
}

ISR(USART_RXC_vect)			//  UART Rx ISR
{
	PushFifoFast(&uart_fifo, UDR);	// empfangenes Zeichen in den uart_fifo legen
}

/****************  Echo Thread  *****************************/
int UartEchoThread(void *vm)	// Threadfunktion, sendet empfangene Zeichen
{
	Mutex* m = (Mutex*)vm;
	
	while(1)
	{
		char c = UartGetChar();	// warten auf ein Zeichen
		LockMutex(m);			// Mutex-Variable sperren
		
		
		UartPutChar(c);			// und das Zeichen ausgeben
		
		if(c == 8)				// DEL wurde gedrckt -> Zeichen lschen
		{
			UartPutChar(' ');
			UartPutChar(8);
		}
		
		ReleaseMutex(m);		// Mutex-Variable wieder freigeben
	}
	
	return 0;
}


/*****************  Thread fr die Demo  *********************/
int UartDemoThread(void *vm)	// Threadfunktion, lsst ein X in der zweiten Zeile des Terminals wandern
{
	Mutex* m = (Mutex*)vm; 		// Zeiger auf Uart-Mutex erzeugen
	static unsigned char pos = 1;
	
	while(1)
	{
		LockMutex(m);		// Mutex-Variable sperren
		SaveCursor();		// Cursorposition speichern
		
		GotoXY(pos++, 2);	// altes 'X' lchen und position erhhen
		UartPutChar(' ');
		Sleep(2);			// Thread fr 100ms deaktivieren
		
		if(pos>80)			// Terminal ist 80 Spalten breit, danach wieder bei 1 anfangen
			pos = 1;
		
		GotoXY(pos, 2);		// 'X' an der neuen Position schreiben
		UartPutChar('X');
		
		RestoreCursor();	// Cursorposition wiederherstellen
		ReleaseMutex(m);	// Mutex-Variable wieder freigeben
		
		Sleep(11);			// Thread fr 110ms deaktivieren
	}
	
	return 0;
}

/**************  Standard-Callback Prozedur  *******************/
void DefProc(int msg, int param)// neue Standard-Prozedur
{
	static unsigned char pos = 1;
	
	switch(msg)
	{
		case MSG_TIMER:		// bei SetTimer als msg angegeben
			LockMutex((Mutex*)param);	// Dasselbe machen wie UartDemoThread, aber dritte Zeile nutzen
			SaveCursor();		// Cursorposition speichern
			
			GotoXY(pos++, 3);
			UartPutChar(' ');
			
			if(pos>80)
				pos = 1;
			
			GotoXY(pos, 3);
			UartPutChar('X');
			
			RestoreCursor();	// Cursorposition wiederherstellen
			ReleaseMutex((Mutex*)param);
			break;
	}
}

/********************  Hauptprogramm  ******************************/
int main(void)
{
	static char uart_echo_thread_stack[128], uart_demo_thread_stack[128];
	THREAD uart_echo_thread, uart_demo_thread;
	
	Mutex uart_mutex;		// Mutex-Variable anlegen und initialisieren
	InitMutex(&uart_mutex);
	
	InitFifo(&uart_fifo, uart_fifo_buffer, sizeof(uart_fifo_buffer));	// UART-FIFO initialisieren
	InitUart(BAUD(19200));	// UART initialisieren
	UartPutChar(12);		// Inhalt im Terminal lschen und Cursor auf 1:1 setzen
	UartPutStr("------------- AVR OS UART Example -------------");
	GotoXY(1, 5); // Cursor an Anfang Zeile 5 setzen
	
	CreateThread(&uart_echo_thread, uart_echo_thread_stack, sizeof(uart_echo_thread_stack), 
			UartEchoThread, (void*)&uart_mutex);	// neue Threads erstellen, zeiger auf
	
	CreateThread(&uart_demo_thread, uart_demo_thread_stack, sizeof(uart_demo_thread_stack), 
			UartDemoThread, (void*)&uart_mutex);	// uart_mutex als Argument
	
	SetSystemDefProc(DefProc);	// Standard-Prozedur ersetzen
	SetTimer(17, NULL, MSG_TIMER, (int)&uart_mutex);	// Timer setzen, der jede 170ms eine Nachricht sendet
	
	while(1)
		DispatchNextMessage();	// Alle Nachrichten abholen und verarbeiten
}
